/*  ___  _
	|   \(_)_ __  ___
	| |) | | '  \/ -_)
	|___/|_|_|_|_\___|

	dialogf.c:
	The dialogf() function and its helpers.

	By Sven Sandberg

	See readme.txt for more information about Dime.
*/
#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <allegro.h>

#ifdef FORTIFY
#include "fortify2.h"
#endif

#include "dime.h"
#include "dime/dimeint.h"



/* Custom callback. */
void (*dialogf_callback)(DIALOG *d) = NULL;

/* The dialog. */
DIALOG *_dime_dialog = NULL;

/* Internal complicated stuff. */
int _dialogf_final_pass, _dialogf_pass_index;

static int draw_background = TRUE;



/* d_background_proc:
	Dialog procedure for drawing the background. Also takes care of yielding
	timeslices if `dime_yield' is set, and of calling `dialogf_callback()' if
	it's not NULL.
*/
static int d_background_proc(int msg, DIALOG *d, int c)
{
	if (msg == MSG_DRAW) {
		/* Not drawing the background is a hack to avoid too much flickering
			when we know we actually don't need to draw anything at all. */
		if (draw_background)
			draw_empty_button(_dime_bmp, d->x, d->y, d->x+d->w-1, d->y+d->h-1,
									_dime_background_color,
									_dime_light_highlight_color,
									_dime_highlight_color,
									_dime_shadow_color, _dime_dark_shadow_color);
		return D_O_K;
	}
	else if (msg == MSG_IDLE) {
		if (dime_yield)
			yield_timeslice();
		if (dialogf_callback)
			dialogf_callback(_dime_dialog);

		return D_O_K;
	}
	else
		return d_box_proc(msg, d, c);
}



/* d_ok_button_proc:
*/
static int d_ok_button_proc(int msg, DIALOG *d, int c)
{
	if (msg == MSG_XCHAR)
		if (((c >> 8) == KEY_ENTER) ||
			 (((c >> 8) == KEY_O) && (key_shifts & KB_ALT_FLAG)))
			return D_CLOSE | D_USED_CHAR;

	return d_dime_button_proc(msg, d, c);
}



/* d_cancel_button_proc:
*/
static int d_cancel_button_proc(int msg, DIALOG *d, int c)
{
	if (msg == MSG_XCHAR)
		if (((c >> 8) == KEY_ESC) ||
			 (((c >> 8) == KEY_C) && (key_shifts & KB_ALT_FLAG)))
			return D_CLOSE | D_USED_CHAR;

	return d_dime_button_proc(msg, d, c);
}



/* d_free_multiline_text_proc:
	Dummy function needed to inform the program that the text must be freed
	after it has been used.
*/
static int d_free_multiline_text_proc(int msg, DIALOG *d, int c)
{
	return d_multiline_text_proc(msg, d, c);
}



/* dialogf_create_desc:
	Creates a description dialog object.
*/
void dialogf_create_desc(DIALOG *d, int y, char *desc)
{
	d->proc = d_multiline_text_proc;
	multiline_text_size(_dime_font, desc, &d->w, &d->h);
	d->x = -dime_window_description_spacing - d->w;
	d->y = y;
	d->fg = _dime_text_color;
	d->bg = _dime_background_color;
	d->key = 0;
	d->flags = 0;
	d->d1 = ALIGN_RIGHT;
	d->d2 = 0;
	d->dp = desc;
	d->dp2 = _dime_font;
	d->dp3 = NULL;
}



/* my_strcmp:
	Returns NULL if the first strlen(name) characters of string are different
	from name or not followed by '[', otherwise returns a pointer to the end
	of `name' within `string' otherwise. `string' should be in the current
	Unicode format, and `name' in pure 7 bit ascii.
*/
static char *my_strcmp(char *string, char *name)
{
	int c1, c2;

	while (1) {
		c1 = ugetx(&string);
		c2 = *name++;
		if (c2 == 0) {
			if (c1 == '[')
				return string;
			else
				return NULL;
		}
		if (c1 != c2)
			return NULL;
	}
}



/* dialogf_start:
	Sets up a `DIALOGF_DIALOG *' struct ready to be used by `dialogf_run()'
*/
DIALOGF_DIALOG *dialogf_start(char *title, int x_pos, int y_pos, int edit_w,
										char *format, dime_va_list_p args)
{
	char arg_buf[4096], desc_buf[4096], *arg_buf_p, *desc_buf_p;
	char *t, *format_end;
	int i, c;
	int more_passes, need_final_pass;
	int found;
	_DIALOGF_FORMAT *df;
	int num_dialogs;
	int num_descs;
	int x, w, x_diff;
	int max_desc_w = 0, title_w, dialog_w, dialog_x;
	int top, bottom, y_diff;
	int object_top, min_y, max_y, dialog_h;
	DIALOGF_DIALOG *dd;
	DIALOG *d;
	int ok_index;
	int did_create;
	int fi;

	int font_h, title_h;

	DIALOG *prev_dime_dialog;
	int prev_dialogf_final_pass, prev_dialogf_pass_index;

	/* Allocate the dialog info and setup global stuff. */
	dd = DIME_MALLOC(1, DIALOGF_DIALOG);

	/* todo: shouldn't this be done separately in each dialogf_[run|start]()
		function? */
	dd->old_gui_fg_color = gui_fg_color;
	dd->old_gui_mg_color = gui_mg_color;
	dd->old_gui_bg_color = gui_bg_color;

	dime_setup_global_stuff();

	font_h = text_height(_dime_font);
	title_h = text_height(_dime_title_font);
	title_w = text_length(_dime_title_font, title);

	/*
	 * Count number of DIALOGs to allocate, check validity of format string
	 * and find width of description texts.
	 */

	t = format;
	w = 0;
	dd->num_formats = 0;
	num_dialogs = 0;
	num_descs = 0;
	for (c = ugetx(&t); c; c = ugetx(&t)) {
		if (c == '%') {
			if (ugetc(t) == '%') {
				c = ugetx(&t);
				w += dime_char_width(_dime_font, '%');
			}
			else {
				dd->num_formats++;

				if (w > max_desc_w)
					max_desc_w = w;
				w = 0;

				/* The special "chain" format must be built-in. */
				/* no, it doesn't need to be, if we split this function into
					dialogf_count() and dialogf_create() */
// todo:finish
//          format_end = my_strcmp(t, "chain");
//          if (format_end == NULL) {
					found = FALSE;
					for (df = _dialogf_formats; df != NULL; df = df->next) {
						format_end = my_strcmp(t, df->specifier);
						if (format_end != NULL) {
							found = TRUE;
							break;
						}
					}
//          }
				DIME_ASSERT(found, ("Invalid format specifier (or specifier not "
										  "followed by '['): %s", t));
				t = format_end;

				arg_buf_p = arg_buf;
				found = FALSE;
				/* See if argument list is ok. */
				for (c = ugetx(&format_end); c; c = ugetx(&format_end)) {
					if (c == '%') {
						c = ugetx(&format_end);
						DIME_ASSERT(c != 0, ("Unexpected end of format string."));
					}
					else if (c == ']') {
						found = TRUE;
						break;
					}
					arg_buf_p += usetc(arg_buf_p, c);
				}
				DIME_ASSERT(found, ("Format string terminated in the middle of "
										  "argument list: %s", t));
				t = format_end;
				usetc(arg_buf_p, 0);
				num_dialogs += df->count_dialog(arg_buf);
				if (!df->handle_desc)
					num_descs++;
			}
		}
		else if (c == '\n') {
			if (w > max_desc_w)
				max_desc_w = w;
			w = 0;
		}
		else
			w += dime_char_width(_dime_font, c);
	}
	DIME_ASSERT(w == 0,
					("Garbage after last specifier in format string: %s",
					 format));

	/*
	 * Construct the dialog.
	 */
	dd->dialog = DIME_MALLOC(5 + num_descs + num_dialogs, DIALOG);
	prev_dime_dialog = _dime_dialog;
	_dime_dialog = dd->dialog;

	dd->objects = DIME_MALLOC(dd->num_formats, DIALOGF_OBJECT);
	for (i = 0; i < dd->num_formats; i++)
		dd->objects[i].again = TRUE;

	/* This is a bit on the complex side: Some objects need to be reconfigured
		when the other objects have been created. Therefore we loop as long as
		there exist objects that request more passes. An object can tell
		`dialogf()' that it wants to be so by returning nonzero from the
		`create_dialog()' function.
		An example of such an object is "%line", which needs to know the total
		width of the dialog so that it can fill up the whole dialog width in
		a nice way.
	*/
	prev_dialogf_final_pass = _dialogf_final_pass;
	prev_dialogf_pass_index = _dialogf_pass_index;
	_dialogf_final_pass = FALSE;
	_dialogf_pass_index = 0;
	need_final_pass = FALSE;
	do {
		more_passes = FALSE;
		d = dd->dialog + 2;

		t = format;
		fi = 0;
		if (_dialogf_pass_index == 0)
			top = 0;
		else
			top = dd->dialog[0].y + dime_window_border_thickness + title_h +
					dime_window_title_spacing;
		bottom = top;
		desc_buf_p = desc_buf;
		for (c = ugetx(&t); c; c = ugetx(&t)) {
			if (c == '%') {
				if (ugetc(t) == '%') {
					c = ugetx(&t);
					desc_buf_p += usetc(desc_buf_p, c);
				}
				/* Found an object! */
				else {
					object_top = bottom;

					/* Get description text for the object. */
					usetc(desc_buf_p, 0);
					desc_buf_p = desc_buf;

					/* Find the type and advance the t pointer. */
					if (_dialogf_pass_index == 0) {
						for (df = _dialogf_formats; df != NULL; df = df->next) {
							format_end = my_strcmp(t, df->specifier);
							if (format_end != NULL)
								break;
						}
						dd->objects[fi].type = df;
						t = format_end;
					}
					else {
						df = dd->objects[fi].type;
						t = my_strcmp(t, df->specifier);
					}

					/* Create description object. */
					if (!df->handle_desc) {
						if (_dialogf_pass_index == 0) {
							dialogf_create_desc(d, bottom, ustrdup(desc_buf));
							d->proc = d_free_multiline_text_proc;
						}
						d++;
					}

					/* Extract argument list. */
					arg_buf_p = arg_buf;
					for (c = ugetx(&t); c; c = ugetx(&t)) {
						if (c == '%') {
							c = ugetx(&t);
							if (c != ']') {
								DIME_ASSERT(c != 0,
												("Format string ended unexpectedly "
												 "inside an escape sequence "
												 "in a format modifier."));
								arg_buf_p += usetc(arg_buf_p, '%');
							}
						}
						else if (c == ']')
							break;
						arg_buf_p += usetc(arg_buf_p, c);
					}
					DIME_ASSERT(c == ']', ("Format string ended unexpectedly "
												  "inside a format modifier."));
					usetc(arg_buf_p, 0);

					/* Create object. */
					if (_dialogf_pass_index == 0) {
						dd->objects[fi].num_dialogs = df->count_dialog(arg_buf);
						dd->objects[fi].dialogs = d;
						for (i = 0; i < dd->objects[fi].num_dialogs; i++) {
							d[i].proc = d_yield_proc;
							d[i].x = 0;
							d[i].y = object_top;
							d[i].w = edit_w;
							d[i].h = font_h;
							d[i].fg = d[i].bg = 0;
							d[i].key = 0;
							d[i].flags = 0;
							d[i].d1 = d[i].d2 = 0;
							d[i].dp = d[i].dp2 = d[i].dp3 = NULL;
						}
					}

					if (dd->objects[fi].num_dialogs &&
						 ((dd->objects[fi].again & 1) ||
						  ((dd->objects[fi].again & 2) && _dialogf_final_pass))) {

						did_create = TRUE;
						dd->objects[fi].again = df->create_dialog(d, desc_buf,
																				arg_buf, args);
						if (dd->objects[fi].again & 1)
							more_passes = TRUE;
						if (dd->objects[fi].again & 2)
							need_final_pass = TRUE;
					}
					else
						did_create = FALSE;

					/* Align objects vertically. */
					if (dd->objects[fi].num_dialogs) {
						min_y = INT_MAX;
						max_y = -INT_MAX;

						for (i = 0; i < dd->objects[fi].num_dialogs; i++) {
							if (d[i].y < min_y)
								min_y = d[i].y;
							if (d[i].y + d[i].h > max_y)
								max_y = d[i].y + d[i].h;
						}

						if (df->handle_desc) {
							y_diff = object_top - min_y;
							for (i = 0; i < dd->objects[fi].num_dialogs; i++)
								d[i].y += y_diff;
							bottom = object_top + max_y - min_y;
						}
						else {
							if (d[-1].h > max_y - min_y) {
								d[-1].y = object_top;
								y_diff = object_top +
											(d[-1].h - (max_y-min_y)) / 2 -
											min_y;
								for (i = 0; i < dd->objects[fi].num_dialogs; i++)
									d[i].y += y_diff;

								bottom = object_top + d[-1].h;
							}
							else {
								d[-1].y = object_top + ((max_y-min_y) - d[-1].h) / 2;
								y_diff = object_top - min_y;
								for (i = 0; i < dd->objects[fi].num_dialogs; i++)
									d[i].y += y_diff;

								bottom = object_top + max_y - min_y;
							}
						}

						if (_dialogf_pass_index == 0) {
							dd->objects[fi].bottom = bottom - top;
						}
						else if (bottom - top != dd->objects[fi].bottom) {
							/* Move the rest of the objects down accordingly. */
							y_diff = (bottom - top) - dd->objects[fi].bottom;
							for (i = dd->objects[fi].num_dialogs;
								  d[i].proc != NULL;
								  i++)
								d[i].y += y_diff;
							for (i = fi + 1; i < dd->num_formats; i++)
								dd->objects[i].bottom += bottom - top -
																 dd->objects[fi].bottom;
							dd->objects[fi].bottom = bottom - top;
						}
					}

					/* Move on to next object. */
					bottom += dime_window_line_spacing;
					d += dd->objects[fi].num_dialogs;
					fi++;
				}
			}
			/* A normal character. */
			else {
				desc_buf_p += usetc(desc_buf_p, c);
			}
		}

		/* Set first non-generated object to NULL. */
		d->proc = NULL;

		/* Fixup alignment of all objects. */
		bottom -= dime_window_line_spacing;
		dialog_h = dime_window_border_thickness +
					  title_h +
					  dime_window_title_spacing +
					  bottom - top +
					  dime_window_button_spacing +
					  dime_window_button_h +
					  dime_window_border_thickness;

		dialog_x = dd->dialog[2].x;
		dialog_w = dd->dialog[2].x + dd->dialog[2].w;
		for (d = dd->dialog + 2; d->proc != NULL; d++) {
			x = d->x;
			if (x < dialog_x)
				dialog_x = x;
			w = d->x + d->w;
			if (w > dialog_w)
				dialog_w = w;
		}
		dialog_w -= dialog_x;
		if (title_w > dialog_w)
			dialog_w = title_w;
		if (dime_window_button_w * 2 + dime_window_between_button_spacing >
			dialog_w)
			dialog_w = dime_window_button_w * 2 +
						  dime_window_between_button_spacing;
		dialog_w += 2 * dime_window_border_thickness;

		if (x_pos == ALIGN_CENTRE)
			x_pos = (_dime_screen_w - dialog_w) / 2;
		else if (x_pos == ALIGN_RIGHT)
			x_pos = _dime_screen_w - dialog_w;
		x_diff = x_pos + dime_window_border_thickness - dialog_x;

		if (y_pos == ALIGN_CENTRE)
			y_pos = (_dime_screen_h - dialog_h) / 2;
		else if (y_pos == ALIGN_BOTTOM)
			y_pos = (_dime_screen_h - dialog_h);
		y_diff = y_pos + dime_window_border_thickness +
					title_h + dime_window_title_spacing - top;

		for (d = dd->dialog + 2; d->proc != NULL; d++) {
			d->x += x_diff;
			d->y += y_diff;
		}

		/* Create the rest of the objects. */

		/* Background. */
		dd->dialog->x = x_pos;
		dd->dialog->y = y_pos;
		dd->dialog->w = dialog_w;
		dd->dialog->h = dialog_h;
		if (_dialogf_pass_index == 0) {
			dd->dialog->proc = d_background_proc;
			dd->dialog->fg =  dd->dialog->bg = 0;
			dd->dialog->key = 0;
			dd->dialog->flags = 0;
			dd->dialog->d1 = dd->dialog->d2 = 0;
			dd->dialog->dp = dd->dialog->dp2 = dd->dialog->dp3 = NULL;
		}

		/* Dialog title. */
		dd->dialog[1].x = x_pos + dime_window_border_thickness;
		dd->dialog[1].y = y_pos + dime_window_border_thickness;
		if (_dialogf_pass_index == 0) {
			dd->dialog[1].proc = d_text_proc;
			dd->dialog[1].w = text_length(_dime_font, title);
			dd->dialog[1].h = title_h;
			dd->dialog[1].fg = _dime_text_color;
			dd->dialog[1].bg = _dime_background_color;
			dd->dialog[1].key = 0;
			dd->dialog[1].flags = 0;
			dd->dialog[1].d1 = dd->dialog[1].d2 = 0;
			dd->dialog[1].dp = title;
			dd->dialog[1].dp2 = _dime_title_font;
			dd->dialog[1].dp3 = NULL;
		}

		/* OK button. */
		d->x = x_pos + dime_window_border_thickness;
		d->y = y_pos + dime_window_border_thickness +
				 title_h + dime_window_title_spacing +
				 bottom - top + dime_window_button_spacing;
		ok_index = d - dd->dialog;
		d->proc = d_ok_button_proc;
		if (_dialogf_pass_index == 0) {
			d->h = dime_window_button_h;
			d->w = dime_window_button_w;
			d->fg = _dime_text_color;
			d->bg = _dime_button_color;
			d->key = 'o';
			d->flags = D_EXIT;
			d->d1 = 0;
			d->d2 = 0;
			d->dp = (char *)get_config_text("&OK");
			d->dp2 = NULL;
			d->dp3 = NULL;
		}
		d++;

		/* Cancel button. */
		d->x = x_pos + dime_window_border_thickness +
				 dime_window_button_w + dime_window_between_button_spacing;
		d->y = y_pos + dime_window_border_thickness + title_h +
				 dime_window_title_spacing + bottom - top +
				 dime_window_button_spacing;
		if (_dialogf_pass_index == 0) {
			d->proc = d_cancel_button_proc;
			d->h = dime_window_button_h;
			d->w = dime_window_button_w;
			d->fg = _dime_text_color;
			d->bg = _dime_button_color;
			d->key = 'c';
			d->flags = D_EXIT;
			d->d1 = 0;
			d->d2 = 0;
			d->dp = (char *)get_config_text("&Cancel");
			d->dp2 = NULL;
			d->dp3 = NULL;
			d++;

			/* NULL-terminating object. */
			d->proc = NULL;
		}

		if (_dialogf_final_pass)
			break;

		if (need_final_pass && !more_passes) {
			_dialogf_final_pass = TRUE;
			more_passes = TRUE;
		}
		_dialogf_pass_index++;
	} while (more_passes);

	/* Reset global variables. */
	_dime_dialog = prev_dime_dialog;
	_dialogf_final_pass = prev_dialogf_final_pass;
	_dialogf_pass_index = prev_dialogf_pass_index;

	return dd;
}



/* dialogf_run:
	Run a dialog once it has been set up by `dialogf_start()'.
*/
int dialogf_run(DIALOGF_DIALOG *dd)
{
	BITMAP *background_bmp = NULL;
	int ret;
	int fi;
	int i;
	int ok_index;
	DIALOG *prev_dime_dialog;
	int do_clean_screen;

	prev_dime_dialog = _dime_dialog;
	_dime_dialog = dd->dialog;

	if (dime_bmp == NULL)
		scroll_screen(0, 0);

	/* Store background. */
	if (dime_clean_screen) {
		background_bmp = create_bitmap(dd->dialog[0].w, dd->dialog[0].h);
		DIME_ASSERT(background_bmp != NULL, ("Out of memory!"));

		scare_mouse_area(dd->dialog[0].x, dd->dialog[0].y,
							  dd->dialog[0].w, dd->dialog[0].h);
		blit(_dime_bmp, background_bmp, dd->dialog[0].x, dd->dialog[0].y,
			  0, 0, dd->dialog[0].w, dd->dialog[0].h);
		unscare_mouse();
	}
	do_clean_screen = dime_clean_screen;

	/* Find index of ok button. */
	for (ok_index = 0;
		  dd->dialog[ok_index].proc != d_ok_button_proc;
		  ok_index++)
		; /* do nothing */

	/* Run. */
	ret = dd->objects[0].dialogs - dd->dialog;
	draw_background = TRUE;
	while (1) {
		ret = do_dialog_find_focus_object(dd->dialog, ret);
		if ((ret >= ok_index) || (ret == -1))
			break;

		fi = -1;
		i = 2;
		while (i <= ret) {
			fi++;
			i += dd->objects[fi].num_dialogs;
			if (!dd->objects[fi].type->handle_desc)
				i++;
		}
		DIME_ASSERT(dd->objects[fi].type->outside_dialog != NULL,
						("Terminating object didn't have an `outside_func()'."));
		if (dd->objects[fi].type->outside_dialog(dd->dialog + ret) == 1) {
			/* Restore background. */
			scare_mouse_area(dd->dialog[0].x, dd->dialog[0].y,
								  dd->dialog[0].w, dd->dialog[0].h);
			blit(background_bmp, _dime_bmp, 0, 0,
										dd->dialog[0].x, dd->dialog[0].y,
										dd->dialog[0].w, dd->dialog[0].h);
			unscare_mouse();
			draw_background = TRUE;
		}
		else
			draw_background = FALSE;
	}

	/* Clean screen. */
	if (do_clean_screen) {
		scare_mouse_area(dd->dialog[0].x, dd->dialog[0].y,
							  dd->dialog[0].w, dd->dialog[0].h);
		blit(background_bmp, _dime_bmp, 0, 0,
									dd->dialog[0].x, dd->dialog[0].y,
									dd->dialog[0].w, dd->dialog[0].h);
		unscare_mouse();
		destroy_bitmap(background_bmp);
	}

	/* Reset global _dime_dialog. */
	_dime_dialog = prev_dime_dialog;

	return (ret == ok_index) ? 1 : 2;
}



/* dialogf_store_results:
	Saves the results after `dialogf_run()' has been called.
*/
void dialogf_store_results(DIALOGF_DIALOG *dd)
{
	int i;
	for (i = 0; i < dd->num_formats; i++)
		if (dd->objects[i].type->store_dialog_result != NULL)
			dd->objects[i].type->store_dialog_result(dd->objects[i].dialogs);
}



/* dialogf_reset_dialog:
	Saves the results after `dialogf_run()' has been called.
*/
void dialogf_reset_dialog(DIALOGF_DIALOG *dd)
{
	int i;
	for (i = 0; i < dd->num_formats; i++)
		if (dd->objects[i].type->reset_dialog != NULL)
			dd->objects[i].type->reset_dialog(dd->objects[i].dialogs);
}



/* dialogf_end:
	Frees the `DIALOGF_DIALOG *' struct and the other stuff that has been
	allocated.
*/
void dialogf_end(DIALOGF_DIALOG *dd)
{
	int i;

	gui_fg_color = dd->old_gui_fg_color;
	gui_mg_color = dd->old_gui_mg_color;
	gui_bg_color = dd->old_gui_bg_color;

	/* Custom clean up function. */
	for (i = 0; i < dd->num_formats; i++)
		if (dd->objects[i].type->destroy_dialog_data != NULL)
			dd->objects[i].type->destroy_dialog_data(dd->objects[i].dialogs);

	/* Free the description texts that we have allocated. */
	for (i = 0; dd->dialog[i].proc != NULL; i++)
		if (dd->dialog[i].proc == d_free_multiline_text_proc)
			free(dd->dialog[i].dp);

	/* Free some more stuff. */
	free(dd->objects);
	free(dd->dialog);
	free(dd);
}



/* _vdialogf:
	Internal function which is almost exactly as `vdialogf()', except that it
	takes a pointer to a va_list rather than a va_list. This is needed in the
	"%dialogf" and "%vdialogf" formats.
*/
int _vdialogf(char *title, int x_pos, int y_pos, int edit_w,
				  char *format, dime_va_list_p args)
{
	int ret;
	DIALOGF_DIALOG *dd;

	dd = dialogf_start(title, x_pos, y_pos, edit_w, format, args);
	dialogf_reset_dialog(dd);
	ret = dialogf_run(dd);
	if (ret == 1)
		dialogf_store_results(dd);
	dialogf_end(dd);

	return ret;
}



/* vdialogf:
	Given a format string, lets the user edit the arguments given using
	Allegro's dialogs. See the documentation for info about the format string.
	Returns TRUE if the user pressed OK and FALSE if he pressed Cancel.
*/
int vdialogf(char *title, int x_pos, int y_pos, int edit_w,
				 char *format, va_list args)
{
	void *void_args = args;
	return _vdialogf(title, x_pos, y_pos, edit_w, format,
						  (dime_va_list_p)&void_args);
}



/* dialogf:
	Wrapper for taking the parameters as variable number of arguments, rather
	than putting them in a list.
*/
int dialogf(char *title, int x, int y, int edit_w,
				char *format, ...)
{
	int ret;
	va_list ap;

	va_start(ap, format);
	ret = vdialogf(title, x, y, edit_w, format, ap);
	va_end(ap);

	return ret;
}