/*  ___  _
	|   \(_)_ __  ___
	| |) | | '  \/ -_)
	|___/|_|_|_|_\___|

	dimegui.c:
	Misc gui helper functions.

	By Sven Sandberg

	See readme.txt for more information about Dime.
*/
#include <allegro.h>
#include <allegro/aintern.h>

#ifdef FORTIFY
#include "fortify2.h"
#endif

#include "dime.h"
#include "dime/dimeint.h"



typedef char *(*GETFUNCPTR)(int, int *);



static int _old_clip_x1, _old_clip_y1, _old_clip_x2, _old_clip_y2, _old_clip;
/* add_clip:
	Adds the given clipping area to the bitmap.
*/
static void add_clip(BITMAP *bmp, int x1, int y1, int x2, int y2)
{
	DIME_ASSERT(bmp != NULL, ("The bitmap was NULL."));

	_old_clip_x1 = bmp->cl;
	_old_clip_y1 = bmp->ct;
	_old_clip_x2 = bmp->cr;
	_old_clip_y2 = bmp->cb;
	_old_clip = bmp->clip;

	if (_old_clip) {
		x1 = MAX(_old_clip_x1, x1);
		x2 = MIN(_old_clip_x2, x2);
		y1 = MAX(_old_clip_y1, y1);
		y2 = MIN(_old_clip_y2, y2);

		if ((x1 > x2) || (y1 > y2))
			x1 = y1 = x2 = y2 = -1;
	}

	set_clip(bmp, x1, y1, x2, y2);
}



/* restore_clip:
	Restores the clipping area after a call to add_clip().
*/
static void restore_clip(BITMAP *bmp)
{
	if (_old_clip)
		set_clip(bmp, _old_clip_x1, _old_clip_y1, _old_clip_x2, _old_clip_y2);
	else
		set_clip(bmp, 0, 0, 0, 0);
}



/* draw_empty_button_frame:
	Draws the frame of a button, without filling it.
*/
void draw_empty_button_frame(BITMAP *bmp,
									  int x1, int y1, int x2, int y2, int face,
									  int xlight, int light, int dark, int xdark)
{
	DIME_ASSERT(bmp != NULL, ("The bitmap was NULL."));

	acquire_bitmap(bmp);

	hline(bmp, x1,   y1,   x2-1, xlight);
	hline(bmp, x1+1, y1+1, x2-2, light);
	hline(bmp, x1+2, y2-1, x2-1, dark);
	hline(bmp, x1+1, y2,   x2,   xdark);
	vline(bmp, x1,   y1+1, y2-1, xlight);
	vline(bmp, x1+1, y1+2, y2-2, light);
	vline(bmp, x2-1, y1+2, y2-2, dark);
	vline(bmp, x2,   y1+1, y2-1, xdark);

	putpixel(bmp, x2,   y1,   face);
	putpixel(bmp, x2-1, y1+1, face);
	putpixel(bmp, x1,   y2,   face);
	putpixel(bmp, x1+1, y2-1, face);

	release_bitmap(bmp);
}



/* draw_empty_button:
	Draws an empty button.
*/
void draw_empty_button(BITMAP *bmp, int x1, int y1, int x2, int y2,
							  int face, int xlight, int light, int dark, int xdark)
{
	DIME_ASSERT(bmp != NULL, ("The bitmap was NULL."));

	acquire_bitmap(bmp);

	draw_empty_button_frame(bmp, x1, y1, x2, y2,
									face, xlight, light, dark, xdark);
	if ((x2 - x1 >= 4) && (y2 - y1 > 4))
		rectfill(bmp, x1 + 2, y1 + 2, x2 - 2, y2 - 2, face);

	release_bitmap(bmp);
}



/* draw_empty_button_down:
	Draws an empty button that is pressed down.
*/
void draw_empty_button_down(BITMAP *bmp, int x1, int y1, int x2, int y2,
									 int face, int xlight, int light, int dark,
									 int xdark)
{
	draw_empty_button(bmp, x1, y1, x2, y2, face, xdark, dark, light, xlight);
}



/* draw_text_button:
	Draws a filled button with text on.
*/
void draw_text_button(BITMAP *bmp, int x1, int y1, int x2, int y2,
							 int face, int textcolor,
							 int xlight, int light, int dark, int xdark,
							 FONT *f, char *text)
{
	FONT *tmp_font;

	DIME_ASSERT(bmp != NULL, ("The bitmap was NULL."));

	/* Temporarily change `font' since gui_strlen() and gui_textout() use
		it. */
	tmp_font = font;
	font = f;

	acquire_bitmap(bmp);

	draw_empty_button(bmp, x1, y1, x2, y2, face, xlight, light, dark, xdark);

	add_clip(bmp, x1+2, y1+2, x2-2, y2-2);
	text_mode(-1);
	gui_textout(bmp, text, (x1+x2+1) / 2,
								  (y1+y2-text_height(f)+1) / 2, textcolor, TRUE);
	restore_clip(bmp);

	font = tmp_font;

	release_bitmap(bmp);
}



/* draw_text_button_down:
	Draws a filled button with text on that is pressed down.
*/
void draw_text_button_down(BITMAP *bmp, int x1, int y1, int x2, int y2,
									int face, int textcolor,
									int xlight, int light, int dark, int xdark,
									FONT *f, char *text)
{
	FONT *tmp_font;

	DIME_ASSERT(bmp != NULL, ("The bitmap was NULL."));

	/* Temporarily change `font' since gui_strlen() and gui_textout() use
		it. */
	tmp_font = font;
	font = f;

	acquire_bitmap(bmp);

	draw_empty_button_down(bmp, x1, y1, x2, y2,
										 face, xlight, light, dark, xdark);

	add_clip(bmp, x1 + 2, y1 + 2, x2 - 2, y2 - 2);
	text_mode(-1);
	gui_textout(bmp, text, (x1+x2+1) / 2 + 2,
								  (y1+y2-text_height(f)+1) / 2 + 2, textcolor, TRUE);
	restore_clip(bmp);

	font = tmp_font;

	release_bitmap(bmp);
}



/* draw_picture_button:
	Draws a filled button with a picture on.
*/
void draw_picture_button(BITMAP *bmp, int x1, int y1, int x2, int y2,
								 int face,int xlight, int light, int dark, int xdark,
								 BITMAP *sprite)
{
	DIME_ASSERT(bmp != NULL, ("The bitmap was NULL."));

	acquire_bitmap(bmp);

	draw_empty_button(bmp, x1, y1, x2, y2, face, xlight, light, dark, xdark);

	add_clip(bmp, x1 + 2, y1 + 2, x2 - 2, y2 - 2);
	draw_sprite(bmp, sprite,
					(x1+x2-sprite->w+1) / 2, (y1+y2-sprite->h+1) / 2);
	restore_clip(bmp);

	release_bitmap(bmp);
}



/* draw_picture_button_down:
	Draws a filled button with a picture on, pressed down.
*/
void draw_picture_button_down(BITMAP *bmp, int x1, int y1, int x2, int y2,
										int face,
										int xlight, int light, int dark, int xdark,
										BITMAP *sprite)
{
	DIME_ASSERT(bmp != NULL, ("The bitmap was NULL."));

	acquire_bitmap(bmp);

	draw_empty_button_down(bmp, x1, y1, x2, y2,
								  face, xlight, light, dark, xdark);

	add_clip(bmp, x1 + 2, y1 + 2, x2 - 2, y2 - 2);
	draw_sprite(bmp, sprite,
					(x1+x2-sprite->w+1) / 2 + 2, (y1+y2-sprite->h+1) / 2 + 2);
	restore_clip(bmp);

	release_bitmap(bmp);
}



/* init_dialog_find_focus_object:
	Like Allegro's `init_dialog()', but focuses the first object after
	`focus_obj' that is willing to take focus, in case `focus_obj' isn't
*/
DIALOG_PLAYER *init_dialog_find_focus_object(DIALOG *dialog, int focus_obj)
{
	DIALOG_PLAYER *player;

	player = init_dialog(dialog, focus_obj);
	if (player->focus_obj != -1)
		dialog[player->focus_obj].flags &= ~D_GOTFOCUS;

	if (focus_obj < 0) {
		player->focus_obj = player->mouse_obj;
		return player;
	}

	while ((dialog[focus_obj].proc != NULL) &&
			 !(SEND_MESSAGE(dialog+focus_obj, MSG_WANTFOCUS, 0) & D_WANTFOCUS))
		focus_obj++;

	if (dialog[focus_obj].proc == NULL)
		player->focus_obj = -1;
	else {
		dialog[focus_obj].flags |= D_GOTFOCUS;
		player->focus_obj = focus_obj;
	}

	return player;
}



/* do_dialog_find_focus_object:
	Like Allegro's `do_dialog()', but focuses the first object after
	`focus_obj' that is willing to take focus, in case `focus_obj' isn't
	focusable.
*/
int do_dialog_find_focus_object(DIALOG *dialog, int focus_obj)
{
	BITMAP *mouse_screen = _mouse_screen;
	int screen_count = _gfx_mode_set_count;
	void *player;

	if (!is_same_bitmap(_mouse_screen, screen))
		show_mouse(screen);

	player = init_dialog_find_focus_object(dialog, focus_obj);
	while (update_dialog(player))
		; /* do nothing */

	if (_gfx_mode_set_count == screen_count)
		show_mouse(mouse_screen);

	return shutdown_dialog(player);
}



/* resize_object:
*/
void resize_object(DIALOG *d, int amount)
{
	d->x -= amount;
	d->y -= amount;
	d->w += amount * 2;
	d->h += amount * 2;
}



/* inwards_bevel:
	Draws a 3d bevel suitable for the checkboxes and editboxes.
*/
void inwards_bevel(DIALOG *d, int ofs, int white_space)
{
	if (white_space)
		rect(DIME_BMP, d->x+2+ofs, d->y+2+ofs,
			  d->x+d->w-3-ofs, d->y+d->h-3-ofs, _dime_edit_field_color);
	hline(DIME_BMP, d->x+ofs, d->y+ofs, d->x+d->w-2-ofs,
			_dime_shadow_color);
	hline(DIME_BMP, d->x+1+ofs, d->y+1+ofs, d->x+d->w-3-ofs,
			_dime_dark_shadow_color);
	vline(DIME_BMP, d->x+ofs, d->y+1+ofs, d->y+d->h-2-ofs,
			_dime_shadow_color);
	vline(DIME_BMP, d->x+1+ofs, d->y+1+ofs, d->y+d->h-3-ofs,
			_dime_dark_shadow_color);
	vline(DIME_BMP, d->x+d->w-1-ofs, d->y+1+ofs, d->y+d->h-2-ofs,
			_dime_edit_field_color);
	hline(DIME_BMP, d->x+ofs, d->y+d->h-1-ofs, d->x+d->w-1-ofs,
			_dime_edit_field_color);
}



/* fill_textout:
*/
void fill_textout(BITMAP *bmp, FONT *f, char *text,
						int x, int y, int w, int fg, int bg)
{
	int text_w, text_h;
	int old_text_mode;

	text_w = text_length(f, text);
	text_h = text_height(f);

	old_text_mode = _textmode;
	text_mode(bg);

	if (is_screen_bitmap(bmp))
		scare_mouse_area(x, y, x+w-1, y+text_h-1);

	textout(bmp, f, text, x, y, fg);
	if (text_w < w)
		rectfill(bmp, x + text_w, y, x + w, y + text_h, bg);

	if (is_screen_bitmap(bmp))
		unscare_mouse();

	text_mode(old_text_mode);
}



/* d_line_proc:
	Simple gui proc for drawing an orthogonal 3d line.
*/
int d_line_proc(int msg, DIALOG *d, int c)
{
	DIME_UNUSED_ARG(c);
	if (msg == MSG_DRAW) {
		if (d->w > d->h) {
			if (d->w > 4) {
				hline(_dime_bmp, d->x, d->y, d->x + d->w - 1,
						_dime_shadow_color);
				hline(_dime_bmp, d->x, d->y + 1, d->x + d->w - 1,
						_dime_light_highlight_color);
			}
		}
		else {
			if (d->h > 4) {
				vline(_dime_bmp, d->x, d->y, d->y + d->h - 1,
						_dime_shadow_color);
				vline(_dime_bmp, d->x + 1, d->y, d->y + d->h - 1,
						_dime_light_highlight_color);
			}
		}
	}
	return D_O_K;
}



/* d_dime_check_proc:
	Dialog proc for 3d-ish check boxes.
*/
int d_dime_check_proc(int msg, DIALOG *d, int c)
{
	if (msg == MSG_DRAW) {
		inwards_bevel(d, 1, TRUE);
		rectfill(_dime_bmp, d->x+4, d->y+4, d->x+d->w-5, d->y+d->h-5,
					_dime_edit_field_color);
		if (d->flags & D_SELECTED) {
			line(_dime_bmp, d->x+4, d->y+4, d->x+d->w-5, d->y+d->h-5, d->fg);
			line(_dime_bmp, d->x+d->w-5, d->y+4, d->x+4, d->y+d->h-5, d->fg);
		}
		if (d->flags & D_GOTFOCUS)
			rect(_dime_bmp, d->x, d->y, d->x+d->w-1, d->y+d->h-1,
				  _dime_border_color);
		else
			rect(_dime_bmp, d->x, d->y, d->x+d->w-1, d->y+d->h-1,
				  _dime_background_color);
		return D_O_K;
	}
	else
		return d_button_proc(msg, d, c);
}



/* dime_draw_scrollable_frame:
 *  Helper function to draw a frame for all objects with vertical scrollbars.
 *  Most of this function is stolen from Allegro.
 */
void dime_draw_scrollable_frame(DIALOG *d, int num_elements)
{
	int handle_y, handle_h;
	int virtual_h;
	int visible_elements;
	int font_h;
	int x1, x2;
	int y1, y2;

	/* draw frame */
	inwards_bevel(d, 0, FALSE);
	font_h = text_height(DIME_FONT);
	virtual_h = font_h * num_elements;

	/* possibly draw scrollbar */
	if (virtual_h > d->h - 4) {
		visible_elements = (d->h - 4) / font_h;
		handle_h = ((d->h - 4) * (d->h - 4) + virtual_h / 2) / virtual_h;
		handle_h = MID(1, handle_h, d->h - 5);

		handle_y = ((d->h - 5 - handle_h) * d->d2 +
						(num_elements - visible_elements) / 2) /
					  (num_elements - visible_elements);

		x1 = d->x + d->w - 10;
		x2 = d->x + d->w - 3;
		y1 = d->y + 2 + handle_y;
		y2 = y1 + handle_h;
		if (handle_y > 0)
			rectfill(screen, x1, d->y + 2, x2, y1, _dime_highlight_color);

		draw_empty_button(screen, x1, y1, x2, y2,
								_dime_button_color,
								_dime_light_highlight_color, _dime_highlight_color,
								_dime_shadow_color, _dime_dark_shadow_color);

		if (y2 < d->y + d->h - 2)
			rectfill(screen, x1, y2 + 1, x2, d->y + d->h - 2,
						_dime_highlight_color);
	}
}



/* dime_draw_listbox:
 *  Helper function to draw a listbox object. Most of this function is stolen
 *  from Allegro.
 */
void dime_draw_listbox(DIALOG *d)
{
	int height, listsize, i, len, bar, x, y, w;
	int fg_color, fg, bg;
	char *sel = d->dp2;
	int font_h;
	char s[1024];

	(*(GETFUNCPTR)d->dp)(-1, &listsize);
	font_h = text_height(DIME_FONT);
	height = (d->h-3) / font_h;
	bar = (listsize > height);
	w = (bar ? d->w-12 : d->w-3);
	fg_color = (d->flags & D_DISABLED) ? gui_mg_color : d->fg;

	/* draw box contents */
	for (i=0; i<height; i++) {
		if (d->d2+i < listsize) {
			if (d->d2+i == d->d1) {
				fg = d->bg;
				bg = fg_color;
			}
			else if ((sel) && (sel[d->d2+i])) {
				fg = d->bg;
				bg = gui_mg_color;
			}
			else {
				fg = fg_color;
				bg = d->bg;
			}
			usetc(s, 0);
			ustrncat(s, (*(GETFUNCPTR)d->dp)(i+d->d2, NULL),
						sizeof(s)-ucwidth(0));
			x = d->x + 2;
			y = d->y + 2 + i*font_h;
			text_mode(bg);
			rectfill(screen, x, y, x+7, y+font_h-1, bg);
			x += 8;
			len = ustrlen(s);
			while (text_length(DIME_FONT, s) >= MAX(d->w - (bar?20:10), 1)) {
				len--;
				usetat(s, len, 0);
			}
			textout(screen, DIME_FONT, s, x, y, fg);
			x += text_length(DIME_FONT, s);
			if (x <= d->x+w)
				rectfill(screen, x, y, d->x+w, y+font_h-1, bg);
		}
		else {
			break;
			rectfill(screen, d->x+2, d->y+2+i*font_h,
								  d->x+w, d->y+1+(i+1)*font_h, d->bg);
		}
	}

	if (d->y+2+i*font_h <= d->y+d->h-3)
		rectfill(screen, d->x+2, d->y+2+i*font_h,
							  d->x+w, d->y+d->h-3, d->bg);

	/* draw frame, maybe with scrollbar */
	dime_draw_scrollable_frame(d, listsize);
}



/* d_dime_text_list_proc:
	Dialog procedure that makes Allegro's listbox more 3d'ish.
*/
int d_dime_text_list_proc(int msg, DIALOG *d, int c)
{
	FONT *old_font;
	int ret;

	old_font = font;
	font = DIME_FONT;

	if (msg == MSG_DRAW) {
		dime_draw_listbox(d);
		ret = D_O_K;
	}
	else
		ret = d_text_list_proc(msg, d, c);

	font = old_font;
	return ret;
}



/* d_dime_edit_proc:
	Dialog procedure that makes Allegro's editbox look more 3d'ish.
*/
int d_dime_edit_proc(int msg, DIALOG *d, int c)
{
	int ret;
	FONT *old_font;
	static int num_resizings = 0;

	old_font = font;
	font = DIME_FONT;

	if (num_resizings == 0)
		resize_object(d, -3);
	num_resizings++;

	ret = d_edit_proc(msg, d, c);

	num_resizings--;
	if (num_resizings == 0) {
		resize_object(d, 3);
		if (msg == MSG_DRAW)
			inwards_bevel(d, 0, TRUE);
	}

	font = old_font;

	return ret;
}



#if 0
/* Not finished */

/* d_dime_edit_proc_:
	An editable text object (the dp field points to an `EDITBOX' struct).
	When it has the input focus (obtained by clicking on it with the mouse),
	text can be typed into this object.
*/
int d_dime_edit_proc_(int msg, DIALOG *d, int c)
{
	EDITBOX *e;
	static int ignore_next_uchar = FALSE;
	int f, l, p, w, x, fg, b, scroll;
	char buf[16];
	char *s;
	int rtm;

	e = d->dp;

	/* calculate maximal number of displayable characters */
/*   if (d->d2 == l)  {
		usetc(buf+usetc(buf, ' '), 0);
		x = text_length(font, buf);
	}
	else
		x = 0;

	b = 0;

	for (p=d->d2; p>=0; p--) {
		usetc(buf+usetc(buf, ugetat(s, p)), 0);
		x += text_length(font, buf);
		b++;
		if (x > d->w)
	 break;
	}

	if (x <= d->w) {
		b = l;
		scroll = FALSE;
	}
	else {
		b--;
		scroll = TRUE;
	}
*/
	switch (msg) {

/*      case MSG_START:
	 d->d2 = l;
	 break;*/

		case MSG_DRAW:
	 fg = (d->flags & D_DISABLED) ? gui_mg_color : d->fg;
	 x = 0;

	 if (scroll) {
		 p = d->d2-b+1;
		 b = d->d2;
	 }
	 else
		 p = 0;

	 for (; p<=b; p++) {
		 f = ugetat(s, p);
		 usetc(buf+usetc(buf, (f) ? f : ' '), 0);
		 w = text_length(font, buf);
		 if (x+w > d->w)
			 break;
		 f = ((p == d->d2) && (d->flags & D_GOTFOCUS));
		 rtm = text_mode((f) ? fg : d->bg);
		 textout(screen, font, buf, d->x+x, d->y, (f) ? d->bg : fg);
		 text_mode(rtm);
		 x += w;
	 }
	 if (x < d->w)
		 rectfill(screen, d->x+x, d->y, d->x+d->w-1, d->y+text_height(font)-1, d->bg);
	 break;

		case MSG_CLICK:
	 x = d->x;

	 if (scroll) {
		 p = d->d2-b+1;
		 b = d->d2;
	 }
	 else
		 p = 0;

	 for (; p<b; p++) {
		 usetc(buf+usetc(buf, ugetat(s, p)), 0);
		 x += text_length(font, buf);
		 if (x > gui_mouse_x())
			 break;
	 }
	 d->d2 = MID(0, p, l);
	 scare_mouse();
	 SEND_MESSAGE(d, MSG_DRAW, 0);
	 unscare_mouse();
	 break;

		case MSG_WANTFOCUS:
		case MSG_LOSTFOCUS:
		case MSG_KEY:
	 return D_WANTFOCUS;

		case MSG_CHAR:
			ignore_next_uchar = FALSE;

			if ((c >> 8) == KEY_LEFT) {
				if (d->d2 > 0)
					d->d2--;
			}
			else if ((c >> 8) == KEY_RIGHT) {
				if (d->d2 < l)
					d->d2++;
			}
			else if ((c >> 8) == KEY_HOME) {
				d->d2 = 0;
			}
			else if ((c >> 8) == KEY_END) {
				d->d2 = l;
			}
			else if ((c >> 8) == KEY_DEL) {
				if (d->d2 < l)
					uremove(s, d->d2);
			}
			else if ((c >> 8) == KEY_BACKSPACE) {
				if (d->d2 > 0) {
					d->d2--;
					uremove(s, d->d2);
				}
			}
			else if ((c >> 8) == KEY_ENTER) {
				if (d->flags & D_EXIT) {
					scare_mouse();
					SEND_MESSAGE(d, MSG_DRAW, 0);
					unscare_mouse();
					return D_CLOSE;
				}
				else
					return D_O_K;
			}
			else if ((c >> 8) == KEY_TAB) {
				ignore_next_uchar = TRUE;
				return D_O_K;
			}
			else {
				/* don't process regular keys here: MSG_UCHAR will do that */
				break;
			}
			scare_mouse();
			SEND_MESSAGE(d, MSG_DRAW, 0);
			unscare_mouse();
			return D_USED_CHAR;

		case MSG_UCHAR:
			if ((c >= ' ') && (uisok(c)) && (!ignore_next_uchar)) {
				if (l < d->d1) {
					uinsert(s, d->d2, c);
					d->d2++;

					scare_mouse();
					SEND_MESSAGE(d, MSG_DRAW, 0);
					unscare_mouse();
				}
				return D_USED_CHAR;
			}
			break;
	}

	return D_O_K;
}
#endif



/* d_dime_button_proc:
	The same as d_button_proc(), but it uses my drawing functions, always
	exits when it is clicked and gets clicked when the user pressed enter.
*/
int d_dime_button_proc(int msg, DIALOG *d, int c)
{
	int ofs;

	if (msg == MSG_DRAW) {
		if (d->flags & D_GOTFOCUS) {
			rect(_dime_bmp, d->x, d->y, d->x+d->w-1, d->y+d->h-1, _dime_border_color);
			ofs = 1;
		}
		else
			ofs = 0;

		if (d->flags & D_SELECTED)
			draw_text_button_down(_dime_bmp, d->x+ofs, d->y+ofs,
														d->x+d->w-1-ofs, d->y+d->h-1-ofs,
										 _dime_button_color, d->fg,
										 _dime_light_highlight_color, _dime_highlight_color,
										 _dime_shadow_color, _dime_dark_shadow_color,
										 _dime_button_font, (char *)d->dp);
		else
			draw_text_button(_dime_bmp, d->x+ofs, d->y+ofs,
												 d->x+d->w-1-ofs, d->y+d->h-1-ofs,
								  _dime_button_color, d->fg,
								  _dime_light_highlight_color, _dime_highlight_color,
								  _dime_shadow_color, _dime_dark_shadow_color,
								  _dime_button_font, (char *)d->dp);
		return D_O_K;
	}
	else if ((msg == MSG_CHAR) && (c >> 8 == KEY_ENTER))
		return d_button_proc(msg, d, c) | D_CLOSE | D_USED_CHAR;
	else
		return d_button_proc(msg, d, c);
}



/* multiline_text_size:
	Finds both the width and the height of a string in pixels. The string may
	include one or more '\n' characters which will be treated as newlines.
*/
void multiline_text_size(FONT *f, char *text, int *w, int *h)
{
	char *end;
	int len;
	int this_w;
	int font_h;
	int c;

	font_h = text_height(f);

	if (w != NULL)
		*w = 0;
	if (h != NULL)
		*h = 0;

	end = text;
	c = ugetc(end);
	while (c) {
		len = 0;
		text = end;
		for (c = ugetx(&end); c && (c != '\n'); c = ugetx(&end))
			len++;

		if (w != NULL) {
			this_w = dime_substr_width(f, text, 0, len);
			if (this_w > *w)
				*w = this_w;
		}
		if (h != NULL)
			*h += font_h;
	}
}



/* d_multiline_text_proc:
	Like d_text_proc, but with several lines.
	The d1 field specifies the alignment as one of the constants:
		ALIGN_LEFT (or 0), ALIGN_CENTRE or ALIGN_RIGHT.
*/
int d_multiline_text_proc(int msg, DIALOG *d, int c)
{
	if (msg == MSG_DRAW) {
		int fg = (d->flags & D_DISABLED) ? gui_mg_color : d->fg;
		FONT *f;
		int x, y;
		char *text, *end;
		int len;
		int font_h;

		if (d->dp2)
			f = d->dp2;
		else
			f = font;

		font_h = text_height(f);
		text_mode(d->bg);

		x = d->x;
		y = d->y;
		end = (char *)d->dp;
		/* Draw each line of text. */
		do {
			/* Find end of line. */
			len = 0;
			text = end;
			for (c = ugetx(&end); c && (c != '\n'); c = ugetx(&end))
				len++;

			/* Find position to draw at. */
			if (d->d1 == ALIGN_RIGHT)
				x = d->x + d->w - dime_substr_width(f, text, 0, len);
			else if (d->d1 == ALIGN_CENTRE)
				x = d->x + (d->w - dime_substr_width(f, text, 0, len)) / 2;

			/* Draw. */
			dime_draw_substr(screen, f, text, x, y, fg, 0, len);

			y += font_h;
		} while (c == '\n');
	}

	return D_O_K;
}



/* dime_char_width:
	Returns the width of one single character of a font. (Allegro only has a
	function for strings). This also treats NULL fonts as having all
	characters one pixel wide, which is useful in a couple of cases when I
	want a NULL font to mean that we're writing to a text screen.
*/
int dime_char_width(FONT *f, int c)
{
	char text[16];
	usetc(text + usetc(text, c), 0);
	return text_length(f, text);
}



/* dime_substr_width:
	Returns the width of only part of a string.
*/
int dime_substr_width(FONT *f, char *text, int start, int length)
{
	int i, c;
	int w = 0;

	DIME_ASSERT(text != NULL, ("The text was NULL."));

	text += uoffset(text, start);

	for (i = 0; i < length; i++) {
		c = ugetx(&text);
		if (!c)
			break;
		w += dime_char_width(f, c);
	}

	return w;
}



/* dime_draw_substr:
	Draws part of a string.
*/
void dime_draw_substr(BITMAP *bmp, FONT *f, char *text, int x, int y,
							 int color, int start, int length)
{
	int end_pos = 0;
	char tmp[512];
	int empty_width = uwidth(EMPTY_STRING);
	int i;
	int current_length;

	DIME_ASSERT(bmp != NULL, ("The bitmap was NULL."));
	DIME_ASSERT(f != NULL, ("The font was NULL."));
	DIME_ASSERT(text != NULL, ("The text was NULL."));

	/* I couldn't think of any way to do it both fast and beautiful; this is
		neither :o) The string is copied into a temporary buffer, and then
		textout() is called with this buffer.
	*/
	text += uoffset(text, start);
	for (i = 0; ugetc(text + end_pos) && (i < length); i++) {
		current_length = uwidth(text + end_pos);
		end_pos += current_length;
		if (end_pos > 512 - empty_width) {
			end_pos -= current_length;
			memcpy(tmp, text, end_pos);
			memcpy(tmp + end_pos, EMPTY_STRING, empty_width);
			textout(bmp, f, tmp, x, y, color);
			x += text_length(font, tmp);
			text += end_pos;
			end_pos = current_length;
			break;
		}
	}
	memcpy(tmp, text, end_pos);
	memcpy(tmp + end_pos, EMPTY_STRING, empty_width);
	textout(bmp, f, tmp, x, y, color);

/* We shouldn't use this since it may be forbidden to patch in a char* even
	though we restore it: String constants may be read-only.

	pos = uoffset(text, start);
	memcpy(backup, text + pos, empty_width);
	usetc(text + pos, empty_char);

	textout(bmp, f, text, x, y, color);

	memcpy(text + pos, backup, empty_width);
*/
}