/*  ___  _
	|   \(_)_ __  ___
	| |) | | '  \/ -_)
	|___/|_|_|_|_\___|

	dlists.c:
	Dialogf functions for selecting a string from a list, ie the formats
	%vlist, %list, %datafile, %wvlist, %wlist, %wdatafile.

	By Sven Sandberg

	See readme.txt for more information about Dime.


	char *my_strings[] = { "foo", "bar", "baz" };
	int index;
	dialogf(..., "Select a name%vlist[2]", &index, my_strings, 3);
	dialogf(..., "Select a name%wvlist[3,Select a name]", &index, my_strings, 3);


	int index;
	dialogf(..., "Select a name%list[3,foo;bar;baz]", &index);
	dialogf(..., "Select a name%wlist[3,Select a name,foo;bar;baz]", &index);


	int index;
	DATAFILE *datafile;
	dialogf(..., "Select a picture from the datafile"
					 "%datafile[3,BMP ;RLE ;CMP ;XCMP]",
					 &index, datafile);
	dialogf(..., "Select a picture from the datafile"
					 "%wdatafile[3,Select,BMP ;RLE ;CMP ;XCMP]",
					 &index, datafile);
	dialogf(..., "Select anything from the datafile%datafile[3,]",
					 &index, datafile);
	dialogf(..., "Select anything from the datafile%wdatafile[3,Select,]",
					 &index, datafile);
*/
#include <allegro.h>
#include <allegro/aintern.h>

#ifdef FORTIFY
#include "fortify2.h"
#endif

#include "dime.h"
#include "dime/dimeint.h"



typedef struct DIALOGF_LIST
{
	int *out;            /* Where to store the result. */
	char **strings;      /* The contents of the list. */
	int num_strings;     /* The number of elements in the list. */
	DATAFILE *datafile;  /* Datafile to use for %datafile. */
	int datafile_types[MAX_DATAFILE_TYPES + 1];  /* Types for %datafile.*/
} DIALOGF_LIST;



#define DATAFILE_END(d) (((d)->type == DAT_END) || ((d)->type == AL_ID('i', 'n', 'f', 'o')))



static DIALOGF_LIST *current_list = NULL;



/* d_my_list_proc:
	A Horrible Hack needed because there is no room for custom data in
	`d_text_list_proc()'.
*/
static int d_my_list_proc(int msg, DIALOG *d, int c)
{
	int ret;
	DIALOGF_LIST *prev_list;

	if (d->dp2 != NULL) {
		prev_list = current_list;
		current_list = d->dp2;
		d->dp2 = NULL;
		ret = d_dime_text_list_proc(msg, d, c);
		d->dp2 = current_list;
		current_list = prev_list;
	}
	else
		ret = d_dime_text_list_proc(msg, d, c);

	return ret;
}



/* my_list_getter:
*/
static char *my_list_getter(int index, int *list_size)
{
	if (index == -1) {
		*list_size = current_list->num_strings;
		return NULL;
	}
	else
		return current_list->strings[index];
}



/* read_datafile_types:
	Reads the semicolon-separated list of datafile types from arg_text.
*/
static int read_datafile_types(DIALOGF_LIST *dl, char **arg_text)
{
	int n;
	char buf[4];
	int c;
	int i;

	c = ugetc(*arg_text);

	if ((c == ',') || (c == 0)) {
		dl->datafile_types[0] = 0;
		ugetx(arg_text);
		return (c == ',') ? DIME_ARG_COMMA : DIME_ARG_END;
	}
	else {
		n = 0;
		do {
			DIME_ASSERT(n < MAX_DATAFILE_TYPES,
							("Too many datafile types in %%datafile."));
			for (i = 0; i < 4; i++) {
				c = ugetx(arg_text);
				DIME_ASSERT((c >= ' ') && (c < 256),
								("Invalid character in datafile type list: '%c' "
								 "(ascii %d)", c, c));
				buf[i] = c;
			}
			dl->datafile_types[n] = AL_ID(buf[0], buf[1], buf[2], buf[3]);
			n++;
			c = ugetx(arg_text);
		} while (c == ';');
		DIME_ASSERT((c == ',') || (c == 0),
						("Datafile types must be exactly 4 characters long and "
						 "separated by semicolons."));
		dl->datafile_types[n] = 0;
		return DIME_ARG_READ | ((c == ',') ? DIME_ARG_COMMA : DIME_ARG_END);
	}
}



/* in_datafile_type_list:
	Returns nonzero if d is in the list in types.
*/
static int in_datafile_type_list(DATAFILE *d, int *types)
{
	int i;

	if (!types[0])
		return TRUE;

	for (i = 0; types[i]; i++)
		if (d->type == types[i])
			return TRUE;

	return FALSE;
}



/* get_datafile_names:
	Find the datafile object names which match the given object types.
*/
static void get_datafile_names(DIALOGF_LIST *dl)
{
	int i, num_added;

	/* Find matching datafile items. */
	dl->num_strings = 0;
	for (i = 0; !DATAFILE_END(dl->datafile + i); i++)
		if (in_datafile_type_list(dl->datafile + i, dl->datafile_types))
			dl->num_strings++;

	dl->strings = DIME_MALLOC(dl->num_strings, char *);
	num_added = 0;
	for (i = 0; !DATAFILE_END(dl->datafile + i); i++)
		if (in_datafile_type_list(dl->datafile + i, dl->datafile_types))
			dl->strings[num_added++] =
			 ustrdup(get_datafile_property(dl->datafile + i, DAT_NAME));
}



/* do_create_list:
	Helper to create a list object. All formats that use some kind of lists
	actually create a `char **' array which they then pass over to this
	function.

	d1 = selected item
	dp2 = DIALOGF_LIST
*/
static void do_create_list(DIALOG *d, DIALOGF_LIST *dl, int num_lines)
{
	int font_h = text_height(_dime_font);

	if (num_lines == 0)
		num_lines = MID(2, dl->num_strings,
							 (3 * _dime_screen_h) / (4 * font_h));

	d->proc = d_my_list_proc;
	d->h = 4 + num_lines * text_height(_dime_font);
	d->fg = _dime_text_color;
	d->bg = _dime_edit_field_color;
	d->d1 = MID(0, *dl->out, dl->num_strings - 1);
	d->d2 = MIN(d->d1, dl->num_strings - num_lines);

	d->dp = my_list_getter;
	d->dp2 = dl;
}



/* do_create_wlist:
	Helper to create a wvlist object. All formats that use some kind of
	lists actually create a `char **' which they then pass over to this
	function.

	d1 = selected item
	d2 = number of lines in dialog
	dp = button text (one of the strings)
	dp2 = DIALOGF_LIST
	dp3 = dialog title
*/
static void do_create_wlist(DIALOG *d, DIALOGF_LIST *dl, int num_lines,
									 char *dialog_title)
{
	int font_h;

	if (num_lines == 0) {
		font_h = text_height(_dime_font);
		num_lines = MID(2, dl->num_strings,
							 (3 * _dime_screen_h) / (4 * font_h));
	}

	d->proc = d_dime_button_proc;
	d->h += 6;
	d->fg = _dime_text_color;
	d->bg = _dime_button_color;
	d->flags |= D_EXIT;
	d->d2 = num_lines;
	d->dp2 = dl;
	d->dp3 = dialog_title;
}



/* dialogf_create_vlist:
	%vlist[lines]
	int *out, char **string, int num_strings
*/
int dialogf_create_vlist(DIALOG *d, char *desc, char *arg_text,
								 dime_va_list_p args)
{
	int state;
	DIALOGF_LIST *dl;
	int num_lines;

	DIME_UNUSED_ARG(desc);

	dl = DIME_MALLOC(1, DIALOGF_LIST);
	dl->out = dime_va_arg_p(args, int *);
	dl->strings = dime_va_arg_p(args, char **);
	dl->num_strings = dime_va_arg_p(args, int);

	/* Parse parameters. */
	state = dime_get_int_arg(&arg_text, &num_lines);
	if (!(state & DIME_ARG_READ))
		num_lines = 0;
	DIME_ASSERT(state & DIME_ARG_END,
					("Garbage found at end of format modifier for %%vlist: [%s]",
					 arg_text));

	do_create_list(d, dl, num_lines);

	return 0;
}



/* dialogf_create_list:
	%list[lines,foo;bar;baz]
	int *out
*/
int dialogf_create_list(DIALOG *d, char *desc, char *arg_text,
								dime_va_list_p args)
{
	int state;
	DIALOGF_LIST *dl;
	int num_lines;

	DIME_UNUSED_ARG(desc);

	state = dime_get_int_arg(&arg_text, &num_lines);
	DIME_ASSERT(state & DIME_ARG_COMMA,
					("You must supply a semicolon-separated list of strings as a "
					 "modifier to %%list."));
	if (!(state & DIME_ARG_READ))
		num_lines = 0;

	dl = DIME_MALLOC(1, DIALOGF_LIST);
	dl->out = dime_va_arg_p(args, int *);

	state = dime_get_strlist_arg(&arg_text, &dl->num_strings, &dl->strings);

	DIME_ASSERT(state & DIME_ARG_END, ("Garbage found at end of modifier to "
												  "%%list: [%s]", arg_text));

	do_create_list(d, dl, num_lines);

	return 0;
}



/* dialogf_create_datafile:
	%datafile[lines,formats]
	int *out, DATAFILE *datafile
*/
int dialogf_create_datafile(DIALOG *d, char *desc, char *arg_text,
									 dime_va_list_p args)
{
	int state;
	DIALOGF_LIST *dl;
	int num_lines;

	DIME_UNUSED_ARG(desc);

	state = dime_get_int_arg(&arg_text, &num_lines);
	if (!(state & DIME_ARG_READ))
		num_lines = 0;
	DIME_ASSERT(state & DIME_ARG_COMMA,
					("You must supply a (possibly empty) list of datafile "
					 "formats to %%datafile."));

	dl = DIME_MALLOC(1, DIALOGF_LIST);
	dl->out = dime_va_arg_p(args, int *);
	dl->datafile = dime_va_arg_p(args, DATAFILE *);

	state = read_datafile_types(dl, &arg_text);
	DIME_ASSERT(state & DIME_ARG_END, ("Garbage found at end of modifier to "
												  "%%datafile: [%s]", arg_text));

	get_datafile_names(dl);

	do_create_list(d, dl, num_lines);

	return 0;
}



/* dialogf_create_wvlist:
	%wvlist[lines,title]
	int *out, char **list, int num_items
*/
int dialogf_create_wvlist(DIALOG *d, char *desc, char *arg_text,
								  dime_va_list_p args)
{
	int state;
	int num_lines;
	char dialog_title[1024];
	DIALOGF_LIST *dl;

	DIME_UNUSED_ARG(desc);

	dl = DIME_MALLOC(1, DIALOGF_LIST);
	dl->out = dime_va_arg_p(args, int *);
	dl->strings = dime_va_arg_p(args, char **);
	dl->num_strings = dime_va_arg_p(args, int);

	/* Parse modifiers. */
	state = dime_get_int_arg(&arg_text, &num_lines);
	if (!(state & DIME_ARG_READ))
		num_lines = 0;
	DIME_ASSERT(state & DIME_ARG_COMMA,
					("You must supply a dialog title as a modifier to "
					 "%%wvlist."));
	state = dime_get_string_arg(&arg_text, dialog_title,
										 sizeof(dialog_title));
	DIME_ASSERT(state & DIME_ARG_END,
					("Garbage found at end of format modifier for %%wvlist: [%s]",
					 arg_text));

	do_create_wlist(d, dl, num_lines, ustrdup(dialog_title));

	return 0;
}



/* dialogf_create_wlist:
	%wlist[lines,title,foo;bar;baz]
	int *out
*/
int dialogf_create_wlist(DIALOG *d, char *desc, char *arg_text,
								 dime_va_list_p args)
{
	int state;
	DIALOGF_LIST *dl;
	int num_lines;
	char dialog_title[1024];

	DIME_UNUSED_ARG(desc);

	state = dime_get_int_arg(&arg_text, &num_lines);
	if (!(state & DIME_ARG_READ))
		num_lines = 0;
	DIME_ASSERT(state & DIME_ARG_COMMA,
					("You must supply a dialog title and a list as modifiers to "
					 "%%wlist."));
	state = dime_get_string_arg(&arg_text, dialog_title,
										 sizeof(dialog_title));
	DIME_ASSERT(state & DIME_ARG_COMMA,
					("You must supply a list as a modifier to %%wlist"));

	dl = DIME_MALLOC(1, DIALOGF_LIST);
	dl->out = dime_va_arg_p(args, int *);

	state = dime_get_strlist_arg(&arg_text, &dl->num_strings, &dl->strings);
	DIME_ASSERT(state & DIME_ARG_END, ("Garbage found at end of modifier to "
												  "%%wlist: [%s]", arg_text));

	do_create_wlist(d, dl, num_lines, ustrdup(dialog_title));

	return 0;
}



/* dialogf_create_wdatafile:
	%wdatafile[lines,title,formats]
*/
int dialogf_create_wdatafile(DIALOG *d, char *desc, char *arg_text,
									  dime_va_list_p args)
{
	int state;
	int num_lines;
	char dialog_title[1024];
	DIALOGF_LIST *dl;

	DIME_UNUSED_ARG(desc);

	dl = DIME_MALLOC(1, DIALOGF_LIST);
	dl->out = dime_va_arg_p(args, int *);
	dl->datafile = dime_va_arg_p(args, DATAFILE *);

	/* Parse modifiers. */
	state = dime_get_int_arg(&arg_text, &num_lines);
	if (!(state & DIME_ARG_READ))
		num_lines = 0;
	DIME_ASSERT(state & DIME_ARG_COMMA,
					("You must supply a dialog title and a (possibly empty) "
					 "list of datafile types as modifiers to %%wdatafile."));
	state = dime_get_string_arg(&arg_text, dialog_title,
										 sizeof(dialog_title));
	DIME_ASSERT(state & DIME_ARG_COMMA,
					("You must supply a a (possibly empty) list of datafile "
					 "types as a modifier to %%wdatafile."));
	state = read_datafile_types(dl, &arg_text);
	DIME_ASSERT(state & DIME_ARG_END,
					("Garbage found at end of format modifier for %%wdatafile: "
					 "[%s]",
					 arg_text));

	/* Find matching datafile items. */
	get_datafile_names(dl);

	do_create_wlist(d, dl, num_lines, ustrdup(dialog_title));

	return 0;
}



/* dialogf_outside_wlists:
*/
int dialogf_outside_wlists(DIALOG *d)
{
	char format[1024];
	int w, max_w;
	int i;
	DIALOGF_LIST *dl = (DIALOGF_LIST *)d->dp2;
	char *title = (char *)d->dp3;

	max_w = dime_window_border_thickness * 2 + dime_window_button_w * 2 +
			  dime_window_between_button_spacing;
	w = text_length(_dime_font, title);
	if (w > max_w)
		max_w = w;
	for (i = 0; i < dl->num_strings; i++) {
		w = text_length(_dime_font, dl->strings[i]);
		if (w > max_w)
			max_w = w;
	}
	if (max_w > (_dime_screen_w * 3) / 4)
		max_w = (_dime_screen_w * 3) / 4;

	usprintf(format, "%%vlist[%d]", d->d2);

	if (dialogf(title, ALIGN_CENTRE, ALIGN_CENTRE, max_w,
				  format, &d->d1, dl->strings, dl->num_strings) == 1)
		d->dp = dl->strings[d->d1];

	return 0;
}



/* dialogf_store_lists:
*/
void dialogf_store_lists(DIALOG *d)
{
	*((DIALOGF_LIST *)d->dp2)->out = d->d1;
}



/* dialogf_store_datafiles:
*/
void dialogf_store_datafiles(DIALOG *d)
{
	DIALOGF_LIST *dsl = (DIALOGF_LIST *)d->dp2;
	int pos;
	int i;

	pos = 0;
	for (i = 0; !DATAFILE_END(dsl->datafile + i); i++) {
		if (in_datafile_type_list(dsl->datafile + i, dsl->datafile_types)) {
			if (pos == d->d1) {
				*dsl->out = i;
				return;
			}
			pos++;
		}
	}
	DIME_ASSERT(0, ("bug:notreached"));
}



/* dialogf_reset_lists:
*/
void dialogf_reset_lists(DIALOG *d)
{
	DIALOGF_LIST *dl = (DIALOGF_LIST *)d->dp2;

	d->d1 = *dl->out;
	d->d1 = MID(0, d->d1, dl->num_strings - 1);
}



/* dialogf_reset_datafile:
*/
void dialogf_reset_datafile(DIALOG *d)
{
	DIALOGF_LIST *dsl = (DIALOGF_LIST *)d->dp2;
	int pos, end;
	int i;

	end = MID(0, *dsl->out, dsl->num_strings - 1);
	pos = 0;
	for (i = 0; i < end; i++)
		if (in_datafile_type_list(dsl->datafile + i, dsl->datafile_types))
			pos++;

	d->d1 = pos;
}



/* dialogf_reset_wlists:
*/
void dialogf_reset_wlists(DIALOG *d)
{
	dialogf_reset_lists(d);
	d->dp = ((DIALOGF_LIST *)d->dp2)->strings[d->d1];
}



/* dialogf_reset_wdatafile:
*/
void dialogf_reset_wdatafile(DIALOG *d)
{
	dialogf_reset_datafile(d);
	d->dp = ((DIALOGF_LIST *)d->dp2)->strings[d->d1];
}



/* dialogf_destroy_vlist:
*/
void dialogf_destroy_vlist(DIALOG *d)
{
	free(d->dp2);
}



/* dialogf_destroy_list_and_datafile:
*/
void dialogf_destroy_list_and_datafile(DIALOG *d)
{
	int i;
	DIALOGF_LIST *dl = (DIALOGF_LIST *)d->dp2;

	for (i = 0; i < dl->num_strings; i++)
		free(dl->strings[i]);
	free(dl->strings);
	free(d->dp2);
}



/* dialogf_destroy_wvlist:
*/
void dialogf_destroy_wvlist(DIALOG *d)
{
	dialogf_destroy_vlist(d);
	free(d->dp3);
}



/* dialogf_destroy_wlist_and_wdatafile:
*/
void dialogf_destroy_wlist_and_wdatafile(DIALOG *d)
{
	dialogf_destroy_list_and_datafile(d);
	free(d->dp3);
}